import asyncio
import datetime
import logging

from .application_layer import FrameParser
from .link_layer import InvalidDatasetError, InvalidChecksumError
from .autodetect import autodetect, detect


class Tic:

    @classmethod
    async def create(cls, device_path):
        res = await detect(device_path)
        if not res:
            raise RuntimeError(f'No tic found at {device_path}')
        return cls(*res)

    def __init__(self, link, serial_number=None):
        self.link = link
        self.device_path = link.line.port
        self.serial_number = serial_number
        self.groups = {}
        self.err_count = 0
        self.last_update = None
        self.updated = asyncio.Event()

    @classmethod
    async def discover(cls):
        tics_params = await autodetect()
        return [cls(*p) for p in tics_params]

    async def infinite_update(self):
        async for frame in self.link.frames():
            try:
                self.groups = FrameParser(frame).to_dict()
                self.last_update = datetime.datetime.now()
                self.updated.set()
            except Exception:
                logging.exception('update error')
            self.err_count = self.link.errors



import serial_asyncio


class Line:
    @classmethod
    def historique(cls, port):
        return cls(port, baudrate=1200)

    @classmethod
    def standard(cls, port):
        return cls(port, baudrate=9600)

    def __init__(self, port, baudrate):
        self.port = port
        self.baudrate = baudrate
        self._reader = None
        self._writer = None

    async def __aenter__(self):
        self._reader, self._writer = await serial_asyncio.open_serial_connection(
            url=self.port,
            baudrate=self.baudrate,
            bytesize=serial_asyncio.serial.SEVENBITS,
            parity=serial_asyncio.serial.PARITY_EVEN,
            stopbits=serial_asyncio.serial.STOPBITS_ONE,
            rtscts=1
        )
        return self

    async def __aexit__(self, *args):
        self._writer.close()
        await self._writer.wait_closed()
        self._reader = None
        self._writer = None

    async def readuntil(self, token):
        return await self._reader.readuntil(token)



import asyncio
import logging

from .physical_layer import Line
from .link_layer import Link
from serial.tools.list_ports import comports


async def autodetect():
    tasks = [asyncio.create_task(detect(port.device)) for port in comports()]
    await asyncio.gather(*tasks)
    return [t.result() for t in tasks if t.result()]


async def detect(port):
    for line in (Line.historique(port), Line.standard(port)):
        try:
            link = Link(line)
            frame = await link.frame()
            serial_number = frame.get("ADSC", frame.get("ADCO"))['data']
        except Exception as e:
            logging.info('No tic detected : %s', e)
            continue
        else:
            return link, serial_number
    return None



from unittest import TestCase
from datetime import datetime

from enedis_tic.link_layer import FrameFactory


def data_set(raw, sep='\t'):
    return FrameFactory.data_set(raw, sep)


class DataSetTest(TestCase):
    def test_without_horodate(self):
        actual = data_set('VTIC\t02\tJ')
        expected = {'label': 'VTIC', 'data': '02'}
        self.assertEqual(expected, actual)

    def test_horodate_without_data(self):
        actual = data_set('DATE\tH240107121638\t\tD')
        expected = {'label': 'DATE', 'data': '', 'datetime': datetime(year=2024, month=1, day=7, hour=12, minute=16, second=38)}
        self.assertEqual(expected, actual)

    def test_with_horodatage(self):
        actual = data_set('SMAXSN\tH240107111839\t01610\t:')
        expected = {'label': 'SMAXSN', 'data': '01610', 'datetime': datetime(year=2024, month=1, day=7, hour=11, minute=18, second=39)}
        self.assertEqual(expected, actual)


from datetime import datetime
import logging


START_FRAME = b'\x02'
END_FRAME = b'\x03'


class Link:
    def __init__(self, line):
        self.line = line
        self.errors = 0

    async def frame(self):
        async with self.line:
            await self._skip_truncated_frame()
            return await self._frame()

    async def frames(self):
        async with self.line:
            await self._skip_truncated_frame()
            while True:
                try:
                    yield await self._frame()
                except Exception:
                    logging.exception('Frame error')
                    self.errors += 1

    async def _frame(self):
        raw = await self.line.readuntil(END_FRAME)
        return FrameFactory(raw.decode('ascii')).to_dict()

    async def _skip_truncated_frame(self):
        await self.line.readuntil(END_FRAME)


class FrameFactory:
    def __init__(self, raw):
        assert raw[0] == START_FRAME.decode('ascii')
        assert raw[-1] == END_FRAME.decode('ascii')
        self.raw = raw
        self.sep = '\t' if '\t' in raw else ' '

    def to_dict(self):
        raw = self.raw[2:-2]  # Ignore start frame, start group, end group and end frame
        if not raw:
            raise RuntimeError('Empty Frame')
        values = raw.split('\r\n')
        return {ds['label']: {'data': ds['data'], 'datetime': ds.get('datetime', None)} for ds in [self.data_set(v, self.sep) for v in values]}

    @staticmethod
    def data_set(raw, sep):
        values = raw.split(sep)
        if not (3 <= len(values) <= 4):
            raise InvalidDatasetError(raw)
        SumChecker(raw).ensure_valid()
        res = {'label': values[0], 'data': values[-2]}
        if len(values) == 4:
            timezone = {
                'H': '+0100',
                'E': '+0200'
            }.get(values[1][0].upper())
            res['datetime'] = datetime.strptime(values[1][1:]+timezone, '%y%m%d%H%M%S%z')
        return res


class InvalidDatasetError(Exception):
    def __init__(self, raw_frame):
        super().__init__()
        self.raw_frame = raw_frame


class SumChecker:
    def __init__(self, raw_data_set):
        self.raw_data_set = raw_data_set

    def verify(self):
        return (self.compute(self._payload) == self.given_checksum)\
               or (self.compute(self._older_payload) == self.given_checksum)

    def ensure_valid(self):
        if not self.verify():
            raise InvalidChecksumError(self.raw_data_set)

    @staticmethod
    def compute(payload):
        s1 = sum([ord(c) for c in payload])
        return (s1 & 0x3F) + 0x20

    @property
    def _payload(self):
        return self.raw_data_set[:-1]

    @property
    def _older_payload(self):
        """For devices up to 2013"""
        return self.raw_data_set[:-2]

    @property
    def given_checksum(self):
        return ord(self.raw_data_set[-1])


class InvalidChecksumError(Exception):
    def __init__(self, raw_dataset):
        super().__init__()
        self.raw_dataset = raw_dataset
